#include <stdlib.h>
#include <stdio.h>
#include <seplib.h>
#include "boost/multi_array.hpp"

typedef boost::multi_array<float, 3> float3D;
typedef boost::multi_array<float, 2> float2D;

int main(int argc, char **argv, char **envp){

	/*initialize IO*/
	initpar(argc,argv);

	/* declare variables*/
	int		debug=0;
	float 	vel_water,ot=0,taper_begin,taper_size;
	int 	n1,n2,n3;
	float 	o1,o2,o3;
	float 	d1,d2,d3;
	float 	*data, *model; //arrays that will hold data, residual read in from command line;
	std::shared_ptr<float3D> modelCube; //boost cubes that will hold data, residual, result cubes;
	
	//GET PARAMETERS######################################################################
	/*check command line parameters*/
	getch("debug","d",&debug);
	//vel_water
	if(!getch("vel_water","f",&vel_water)) seperr("could not get vel_water from param\n");
	else if(debug) fprintf(stderr,"vel_water=%f m/s \n",vel_water);
	
	//taper_size
	if(!getch("taper_size","f",&taper_size)) seperr("could not get taper_size from param\n");
	else if(debug) fprintf(stderr,"taper_size=%f sec \n",taper_size);
	
	//ds
	if(!getch("ds","f",&d3)) seperr("could not get ds from param\n");
	else if(debug) fprintf(stderr,"ds=%f meters\n",d3);
	
	//os
	if(!getch("os","f",&o3)) seperr("could not get os from param\n");
	else if(debug) fprintf(stderr,"os=%f meters \n",o3);
	
	//dr
	if(!getch("dr","f",&d2)) seperr("could not get dr from param\n");
	else if(debug) fprintf(stderr,"dr=%f meters \n",d2);
	
	//or
	if(!getch("or","f",&o2)) seperr("could not get or from param\n");
	else if(debug) fprintf(stderr,"or=%f meters\n",o2);
	
	//dt
	if(!getch("dt","f",&d1)) seperr("could not get dt from param\n");
	else if(debug) fprintf(stderr,"dt=%f sec\n",d1);
	
	//ot
	if(!getch("ot","f",&o1)) seperr("could not get dt from param\n"); 
	else if(debug) fprintf(stderr,"ot=%f sec\n",o1);
	
	//wavelet_delay
	if(!getch("taper_begin","f",&taper_begin)) seperr("could not get taper_begin from param\n");
	else if(debug) fprintf(stderr,"taper_begin=%f sec\n",taper_begin);
	
	/*get data dimensions from std in*/
	if(!hetch("n1","d",&n1)) seperr("could not get n1 from file\n"); //nt
	else if(debug) fprintf(stderr,"n1=%d \n",n1);
	if(!hetch("n2","d",&n2)) seperr("could not get n2 from file\n"); //nr
	else if(debug) fprintf(stderr,"n2=%d \n",n2);
	if(!hetch("n3","d",&n3)) seperr("could not get n3 from file\n"); //ns
	else if(debug) fprintf(stderr,"n3=%d \n",n3);
	//####################################################################################
	
	/*read in residual from stdin*/
	data = new float [n1*n2*n3];
	if(!sreed("in",data,sizeof*data*n1*n2*n3)) seperr("could not read data\n");
	else if(debug) fprintf(stderr,"data sreed successful\n");
	
	/*copy to boost array*/
	modelCube.reset(new float3D(boost::extents[n3][n2][n1]));
	for(int i3=0;i3<n3;i3++){
		for(int i2=0;i2<n2;i2++){
			for(int i1=0;i1<n1;i1++){
				int index=i3*n2*n1 + i2*n1 + i1;
				(*modelCube)[i3][i2][i1]=data[index];
			}
		}
	}
	
	int taper_size_index=taper_size/d1;
	int taper_begin_index=taper_begin/d1;
	/*mute direct arrival*/
	float weight;
    for(int i3=0; i3<n3; i3++){
    	float s = o3 + i3*d3;
    	for(int i2=0; i2<n2; i2++){
    		float r = o2 + i2*d2;
    		float dsr = abs(r-s);
    		float t = dsr/vel_water + o1 + taper_begin + taper_size;
    		int it = (t-o1)/d1;
    		//if(it>n1) it=n1;
    		if(debug){fprintf(stderr,"mute shot %d, rec %d, weight=0 before %f (%d). weight=1 after %f (%d), taper_index_size=%d\n",i3,i2,t-taper_size, it-taper_size_index, t, it, taper_size_index);}
    		//weight=1.0/(.5*3.14159);
    		for(int i1=0; i1<it; i1++){
    			if(abs(i1-it)>taper_size_index){
    				weight=0;
    			}
    			else {
    				weight=cos(3.14159*.5*(float)abs(it-i1)/taper_size_index);
    				weight=weight*weight;
    			}
    			if(debug) fprintf(stderr,"time index=%d : weight=%f\n",i1,weight);
    			if(i1<n1) (*modelCube)[i3][i2][i1]=(*modelCube)[i3][i2][i1]*weight;
    		}
    	}
    }
    
    model = new float [n1*n2*n3];
	for(int i3=0;i3<n3;i3++){
		//for each trace
		for(int i2=0;i2<n2;i2++){
			//writ to array
			for(int i1=0;i1<n1;i1++){
				int index=i3*n2*n1 + i2*n1 + i1;
				model[index]=(*modelCube)[i3][i2][i1];
			}
		}
	}
	
	if(!srite("out",model,sizeof*model*n1*n2*n3)) seperr("Could not write out model");
	else if(debug) fprintf(stderr,"model write successful\n");	
}