#include <stdlib.h>
#include <stdio.h>
#include <seplib.h>
#include "boost/multi_array.hpp"

typedef boost::multi_array<float, 3> float3D;
typedef boost::multi_array<float, 2> float2D;

int main(int argc, char **argv, char **envp){

	/*initialize IO*/
	initpar(argc,argv);

	/* declare variables*/
	char 	dataFilename[128];//residual file is std in
	int		mod=0,debug=0;
	int 	n1,n2,n3;
	float 	o1,o2,o3;
	float 	d1,d2,d3;
	float 	*data, *res, *result; //arrays that will hold data, residual read in from command line;
	std::shared_ptr<float3D> dataCube, resCube, resultCube; //boost cubes that will hold data, residual, result cubes;
	std::shared_ptr<float2D> norms, dataDotRes; //boost arrays to hold norm of each trace, dot product result between the each data trace and the matching residual trace
	
	/*check mod parameter*/
	getch("mod","d",&mod);
	getch("debug","d",&debug);
	
	/*get data dimensions from std in*/
	if(!hetch("n1","d",&n1)) seperr("could not get n1 from file\n"); //nt
	if(!hetch("n2","d",&n2)) seperr("could not get n2 from file\n"); //nr
	if(!hetch("n3","d",&n3)) seperr("could not get n3 from file\n"); //ns

	/*read in residual from stdin*/
	res = new float [n1*n2*n3];
	if(!sreed("in",res,sizeof*res*n1*n2*n3)) seperr("could not read residual\n");
	else if(debug) fprintf(stderr,"residual sreed successful\n");
	
	/*read in data from data parameter*/
	data = new float [n1*n2*n3];
	if(!getch("data","s",&dataFilename)) seperr("data file not defined");
 	else if(debug) fprintf(stderr,"data file name: %s\n",dataFilename);
 	
	if(!sreed(dataFilename,data,sizeof*data*n1*n2*n3)) seperr("could not read data. Check the file exists and is the same size as residual file.\n");
	else if(debug) fprintf(stderr,"data sreed successful\n");
	
	//debugging
	if(debug){
		fprintf(stderr,"residual dimensions - nt:%d nr:%d ns:%d \n",n1,n2,n3);
		fprintf(stderr,"mod=%d\n",mod);
	}
	
	/*copy residual and data into respective boost cubes. boost cubes are much easier to index through and behave well with openmp/tbb*/
	/*also calculate norms*/ 
	dataCube.reset(new float3D(boost::extents[n3][n2][n1]));
	resCube.reset(new float3D(boost::extents[n3][n2][n1]));
	norms.reset(new float2D(boost::extents[n3][n2]));
	dataDotRes.reset(new float2D(boost::extents[n3][n2]));
	for(int i3=0;i3<n3;i3++){
		for(int i2=0;i2<n2;i2++){
			(*norms)[i3][i2]=0;
			(*dataDotRes)[i3][i2]=0;
			if(debug){fprintf(stderr,"*********************************\n calculating norm and dataDotRes\n nr:%d ns:%d\n",i2,i3);}
			for(int i1=0;i1<n1;i1++){
				int index=i3*n2*n1 + i2*n1 + i1;
				(*norms)[i3][i2]+=(data[index]*data[index]);
				(*dataDotRes)[i3][i2]+=(data[index]*res[index]);
				(*dataCube)[i3][i2][i1]=data[index];
				(*resCube)[i3][i2][i1]=res[index];
				/*if(debug){
					fprintf(stderr,"nt: %d \n index: %d \n data[index]: %f \n (*norms)[%d][%d]: %f\n",i1,index,data[index],n3,n2,(*norms)[i3][i2]);
					system("PAUSE");
				}*/
			}
			(*norms)[i3][i2]=pow((*norms)[i3][i2],.5);
			if(debug){
					fprintf(stderr,"(*norms)[%d][%d]: %f\n",i3,i2,(*norms)[i3][i2]);
			}
		}
	}
	
	resultCube.reset(new float3D(boost::extents[n3][n2][n1]));
	/*if mod parameter was 0 do res/||d|| */
	if(mod==0){
		//for each shot
		for(int i3=0;i3<n3;i3++){
			//for each trace
			for(int i2=0;i2<n2;i2++){
				//divide each time sample by the norm
				float norm = (*norms)[i3][i2];
				for(int i1=0;i1<n1;i1++){
					 (*resultCube)[i3][i2][i1]=(*resCube)[i3][i2][i1]/norm;
				}
			}
		}
	}		
	/*if mod parameter was not 0 do (res^(t)*dat*dat)/||d||^3 */
	else{
		//for each shot
		for(int i3=0;i3<n3;i3++){
			//for each trace
			for(int i2=0;i2<n2;i2++){
				//divide each sample by the norm cubed
				float norm = pow((*norms)[i3][i2],3);
				float dotResult = (*dataDotRes)[i3][i2];
				for(int i1=0;i1<n1;i1++){
					(*resultCube)[i3][i2][i1]=(*dataCube)[i3][i2][i1]*dotResult/norm;
				}
			}
		}
	}		
	
	
	/*write result to float array then write to stdout*/
	result = new float [n1*n2*n3];
	for(int i3=0;i3<n3;i3++){
		//for each trace
		for(int i2=0;i2<n2;i2++){
			//writ to array
			for(int i1=0;i1<n1;i1++){
				int index=i3*n2*n1 + i2*n1 + i1;
				result[index]=(*resultCube)[i3][i2][i1];
			}
		}
	}
	
	if(!srite("out",result,sizeof*result*n1*n2*n3)) seperr("Could not write out result");
	else fprintf(stderr,"result write successful\n");		
}